/* This module provides log output for the http server.
 *
 * The output may be routed to any of three files:
 *     Error log
 *     Access log
 *     Trace log
 *
 * This module provides output to an application log file.  It formats
 * data via SYS$FAO to facilitate putting timestamps in the control strings.
 *
 *  int tlog_init ( char *log_file_name );
 *  int tlog_initlog ( int type, char * fname );	! type 1 
 *  int tlog_putlog ( int level, char *ctlstr, ... );
 *  int tlog_flush;
 *  int tlog_reopen ( level );
 *
 * Author:	David Jones
 * Revised:	27-AUG-1994		Added tlog_reopen().
 */
#include <stdio.h>
#include <stdarg.h>
#include <descrip.h>
#include <ssdef.h>
#include <string.h>
#include "pthread_np.h"
#include <errno.h>

static FILE *log_file;			/* error messages (type 0) */
static FILE *acc_file;			/* access log (type 1) */
static FILE *trc_file;			/* trace file */
static int log_valid = 0;
static int acc_valid = 0;
static int trc_valid = 0;
static int trace_level = 0;
static struct dsc$descriptor control;	/* control string */
static struct dsc$descriptor output;	/* control string */
static char log_file_spec[256];
static char acc_file_spec[256];
static char trc_file_spec[256];
static pthread_mutex_t log_write;
static int flush_ef;
static long flush_interval[2];
static int flush_enable;
int LIB$GET_EF(), SYS$CLREF(), SYS$SETIMR();
int http_log_level;
#ifdef __DECC
/* Work around problems with fsync() not being ANSI standard */
#ifndef fsync
#define fsync decc$fsync
int decc$fsync(int fd);
#endif
#else
int fsync();
#endif

int tlog_init ( char *log_file_name )
{
    if ( log_valid ) return 3;
    log_file = fopen ( log_file_name, "w",
	"dna=.log",
	"alq=100", "deq=600", /* "bls=8192", */
	"rfm=var", "mrs=300",
	"ctx=rec", "shr=upd");
    strcpy ( log_file_spec, log_file_name );
    if ( log_file ) log_valid = 1;
    if ( !log_valid ) perror ( "Can't open log file" );
    LIB$GET_EF ( &flush_ef );
    SYS$CLREF ( flush_ef );
    flush_enable = 0;
    flush_interval[0] = -600000000; flush_interval[1] = -1;
    SYS$SETIMR ( flush_ef, &flush_interval, 0, 0, 0 );

    /*
     * initialize static fields in descriptors.
     */
    control.dsc$b_dtype = DSC$K_DTYPE_T;
    control.dsc$b_class = DSC$K_CLASS_S;
    output.dsc$b_dtype = DSC$K_DTYPE_T;
    output.dsc$b_class = DSC$K_CLASS_S;
    /*
     * Initialize mutexes and conditions.
     */
    pthread_mutex_init ( &log_write, pthread_mutexattr_default );

    return log_valid;
}
/***********************************************************************/
/* Initialize secondary log files (access log and trace log).
 * If trace log file name is empty, send trace into to error log.
 */

int tlog_initlog ( int level,		/* which file: -1 access, 1 trace */
	char *fname )
{
   if ( level < 0 ) {		/* access log */
       if ( acc_valid ) return 3;
        strcpy ( acc_file_spec, fname );
        acc_file = fopen ( fname, "a",
	    "dna=.log",
	    "alq=100", "deq=600", /* "bls=8192", */
	    /* "rfm=var",*/ "mrs=900",
	    /* "ctx=rec", */ "shr=upd");
	if ( acc_file ) acc_valid = 1;
	return acc_valid;

    } else {
	trace_level = level;
        if ( trc_valid ) return 3;		/* Already selected */
	if ( *fname == '\0' ) fname = log_file_spec;
        strcpy ( trc_file_spec, fname );
	if ( strcmp ( fname, log_file_spec ) == 0 ) {
	    /*
	     * Trace log and error log are same, make log file invalid.
	     */
	    trc_file = log_file;
	    log_valid = 0;
	    trc_valid = 1;
	    return trc_valid;
	}
        trc_file = fopen ( fname, "w",
	    "dna=.log",
	    "alq=100", "deq=600", /* "bls=8192", */
	    "rfm=var", "mrs=1200",
	    "ctx=rec", "shr=upd");
	return trc_valid;
    }
 }

/*
 * Format log output and send to appropriate log file based upon log
 * level argument and trace level set with tlog_initlog;
 *     0	Error, write to trace file and error log.
 *    <0	Normal access, write to access log.
 *    >0	Write to trace file if trace_level greater or equal to level.
 */
int tlog_putlog ( int level, char *ctlstr, ... )
{
    va_list param_list;
#ifdef __ALPHA
    unsigned int __VA_COUNT_BUILTIN(void), arg_count;
    int i, fao_param[32];
#endif
    int status, SYS$FAOL(), length, clen, valid;
    char outbuf[1000];
    FILE *fp;
    va_start(param_list,ctlstr);
    /*
     * See if log file available and synchronize access to static areas.
     */
    if ( !log_valid && !acc_valid && !trc_valid ) return 0;
    if ( (http_log_level == 997) && (level > 1) ) {
	int ts_tcp_stack_used();
	tlog_putlog ( 1, "Stack usage: !SL bytes!/", ts_tcp_stack_used() );
    }
    pthread_mutex_lock ( &log_write );
    for ( clen = 0; ctlstr[clen]; clen++ );
    control.dsc$w_length = clen;
    control.dsc$a_pointer = ctlstr;
    output.dsc$w_length = sizeof(outbuf);
    output.dsc$a_pointer = outbuf;

    length = 0;
    LOCK_C_RTL
#ifdef __ALPHA
    arg_count = __VA_COUNT_BUILTIN() - 1;
    for ( i = 0; i < arg_count; i++ ) fao_param[i] = va_arg(param_list,int);
    LOCK_VMS_RTL
    status = SYS$FAOL ( &control, &length, &output, fao_param );
#else
    LOCK_VMS_RTL
    status = SYS$FAOL ( &control, &length, &output, param_list );
#endif
    UNLOCK_VMS_RTL
    if ( (status&1) == 1 ) {
	if ( level < 0 ) {
	    status = acc_valid ? fwrite(outbuf, length, 1, acc_file) : 1;
	    if ( status != 1 ) {
	        char errtxt[64];
	        sprintf(errtxt, "access log write error (length=%d)", length );
	        perror ( errtxt );
	    }
	} else {
	    status = log_valid && ( 0 == level ) ? 
			fwrite(outbuf, length, 1, log_file) : 1;
	    if ( status != 1 ) {
	        char errtxt[64];
	        sprintf(errtxt, "log file write error (length=%d)", length );
	        perror ( errtxt );
	    }
	    status = trc_valid && ( trace_level >= level ) ?
			 fwrite(outbuf, length, 1, trc_file) : 1;
	    if ( status != 1 ) {
	        char errtxt[64];
	        sprintf(errtxt, "trace file write error (length=%d)", length );
	        perror ( errtxt );
	    }
	}
    } else {
	fprintf(log_file,"FAO error: %d '%s'\n", status, ctlstr );
    }
    UNLOCK_C_RTL
    flush_enable = 1;
    pthread_mutex_unlock ( &log_write );
    return status;
}


int tlog_flush()
{
    if ( log_valid  || acc_valid || trc_valid ) {
	int status, ef_mask, fsync(), SYS$READEF();
	pthread_mutex_lock ( &log_write );
	if ( flush_enable ) {
	    flush_enable = 0;
	    if ( SYS$READEF(flush_ef,&ef_mask) == SS$_WASSET ) {
		SYS$CLREF(flush_ef);
		SYS$SETIMR ( flush_ef, flush_interval, 0, 0, 0 );
	        if ( log_valid ) {
		    LOCK_C_RTL
		    fflush ( log_file );
		    status = fsync ( fileno(log_file) );
		    UNLOCK_C_RTL
		}
	        if ( acc_valid ) {
		    LOCK_C_RTL
		    fflush ( acc_file );
		    status = fsync ( fileno(acc_file) );
		    UNLOCK_C_RTL
		}
	        if ( trc_valid ) {
		    LOCK_C_RTL
		    fflush ( trc_file );
		    status = fsync ( fileno(trc_file) );
		    UNLOCK_C_RTL
		}
	    }
	} else status = 0;
	pthread_mutex_unlock ( &log_write );
	return status;
    }else return 0;
}
/*****************************************************************************/
/* Make new version of open log file. */
tlog_reopen ( int level ) {
{
   if ( level < 0 ) {		/* access log */
        /*
	 * Make new version of access log.
	 */
	pthread_mutex_lock ( &log_write );
       if ( acc_valid ) {
	    LOCK_C_RTL
	    fclose ( acc_file );
            acc_file = fopen ( acc_file_spec, "w",
	        "dna=.log",
	        "alq=100", "deq=600", /* "bls=8192", */
	        /* "rfm=var",*/ "mrs=900",
	        /* "ctx=rec", */ "shr=upd");
	    UNLOCK_C_RTL
	    if ( acc_file ) acc_valid = 1;
	}
	pthread_mutex_unlock ( &log_write );
	return acc_valid;

    } else {
	/*
	 * Make version version of trace log and set logger level.
	 */
	pthread_mutex_lock ( &log_write );
	if ( trc_valid ) {
	    trace_level = level;
	    LOCK_C_RTL
	    fclose ( trc_file );
            trc_file = fopen ( trc_file_spec, "w",
	        "dna=.log",
	        "alq=100", "deq=600", /* "bls=8192", */
	        "rfm=var", "mrs=1200",
	        "ctx=rec", "shr=upd");
	    UNLOCK_C_RTL
	}
	pthread_mutex_unlock ( &log_write );
	return trc_valid;
    }
 }
}
